
/******************************************************************************
* MODULE     : ext_generic.cc
* DESCRIPTION: Produce the generic extension code
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
#include <quantor.hh>
namespace mmx {

/******************************************************************************
* Specializing list of imports to a given type
******************************************************************************/

generic
extender::specialize (generic g, generic to, bool virtual_flag) {
  // Importations should all be of the form
  //   forall(T:Class) f: (A_1,...,A_n)->B == "f";
  // Genericity over more than one type is not yet implemented.
  // If one among A_1,...,A_n,B is of the form F(T),
  // where F(T) involves T, but is not equal to T or Alias T,
  // then a converter F(X)->F(Y) should be supplied for all X and Y
  // for which a converter X->Y exists.

  if (!is_func (g, MMX_FORALL, 2))
    error (g, "Expected a 'forall' type");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g[1], "Expected a typed quantification variable");
  else if (g[1][2] != MMX_CLASS)
    error (g[1][2], "Expected 'Class' type");
  else if (is_func (g[2], MMX_FORALL, 2))
    error (g, "Double quantifications not yet allowed");
  else if (!is_func (g[2], MMX_DEFINE, 2))
    error (g[2], "Expected a declaration");
  else if (!is_func (g[2][1], MMX_TYPE, 2))
    error (g[2], "Expected a declaration");
  else {
    generic qvar= g[1][1];
    generic body= g[2];
    generic ret = substitute (body, qvar, to);
    if (virtual_flag)
      ret= list (ret[0],
		 list (ret[1][0], ret[1][1], Virtual (ret[1][2])),
		 ret[2]);
    return ret;
  }
  return g;
}

list
extender::specialize (list l, generic to, bool virtual_flag) {
  if (nil (l)) return l;
  else return cons (specialize (car (l), to, virtual_flag),
		    specialize (cdr (l), to, virtual_flag));
}

/******************************************************************************
* Specializing list of imports to a given type
******************************************************************************/

generic
extender::type_as_ccpp (generic g, table t) {
  // search the C/C++ name of a Mathemagix type
  // some additional conversion rules may be specified in the table
  if (t -> contains (g)) return t[g];
  else if (type_conv -> contains (g)) return type_conv [g];
  else if (is_func (g, MMX_CROSS, 0)) return MMX_VOID;
  else if (is_func (g, MMX_ALIAS, 1)) return Alias (type_as_ccpp (g[1], t));
  else if (is_list (g)) {
    list l= as_list (g), r;
    while (!nil (l)) {
      r= cons (type_as_ccpp (car (l), t), r);
      l= cdr (l);
    }
    return reverse (r);
  }
  else if ((g == MMX_CROSS) ||
	   (g == MMX_FUNCTION) ||
	   (g == MMX_POINTER) ||
	   (g == MMX_ARRAY) ||
	   (g == MMX_CLASS) ||
	   (g == MMX_BOOL) ||
	   (g == MMX_VOID) ||
	   (g == CCPP_CONST) ||
	   (g == CCPP_VIRTUAL) ||
	   (g == geninst ()))
    return g;
  else {
    int i;
    for (i=1; i<=max_type_args; i++)
      if (g == geninst (i))
	return g;
    warning (g, "Type could not be translated");
    return g;
  }
}

list
extender::type_as_ccpp (list l, table t) {
  if (nil (l)) return l;
  return cons (type_as_ccpp (car (l), t), type_as_ccpp (cdr (l), t));
}

generic
extender::decl_as_ccpp (generic g, bool gen_flag) {
  if (!is_func (g, MMX_DEFINE, 2))
    error (g, "Expected a declaration");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[2]))
    error (g[2], "Expected a symbol");
  else if (!gen_flag)
    return list (g[1][0], g[2], type_as_ccpp (g[1][2]));
  else return list (g[1][0], genmethod (g[2]), type_as_ccpp (g[1][2]));
  return g;
}

list
extender::decl_as_ccpp (list l, bool gen_flag) {
  if (nil (l)) return l;
  else return cons (decl_as_ccpp (car (l), gen_flag),
		    decl_as_ccpp (cdr (l), gen_flag));
}

/******************************************************************************
* Produce the gentype type. This type serves as a base type for
* all dynamic types which have been exported by the user.
******************************************************************************/

static list
arg_list (int start, int end) {
  if (start >= end) return list ();
  else return cons (parameter (start), arg_list (start+1, end));
}

void
extender::gentype_methods (generic g) {
  if (!is_func (g, MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1]))
    error (g[1], "Expected a symbol");
  else if (!is_func (g[2], MMX_FUNCTION, 2))
    error (g[2], "Expected a function type");
  else if (!is_func (g[2][1], MMX_CROSS))
    error (g[2][1], "Expected a cross type");
  else {
    g= name_parameters (g);
    generic name= genmethod (g[1]);
    generic scoped= Scope (gentype_rep, name);
    g= list (g[0], scoped, g[2]);
    generic _error= list (MMX_LESSLESS, symbol ("cerr"),
			  symbol ("\"Fatal error: unsupported method\\n\""));
    generic _exit = list (symbol ("exit"), symbol ("1"));
    generate (Define (g, Begin (_error, _exit)));
  }
}

void
extender::gentype_methods (list l) {
  if (!nil (l)) {
    gentype_methods (car (l));
    gentype_methods (cdr (l));
  }
}

void
extender::produce_gentype () {
  generate (Type (gentype_rep, MMX_CLASS));
  generate (Typedef (gentype, Pointer (gentype_rep)));
  list l= specialize (imports, geninst (), true);
  l= decl_as_ccpp (l, true);
  generic gc= Define (Type (gentype_rep, Inline (Constructor (MMX_VOID))),
		      Begin ());
  generic gd= Define (Type (gentype_rep, Inline (Virtual (Destructor ()))),
		      Begin ());
  l= cons (gc, gd, l);
  generate (make_struct (gentype_rep, l));
  l= specialize (imports, geninst (), false);
  l= decl_as_ccpp (l);
  gentype_methods (l);
}

/******************************************************************************
* Produce the generic instance types, indexed by numbers 1, 2, ....
* These types are instances of dynamic types. curinst(i) holds
* the current type of geninst(i).
******************************************************************************/

static generic
converter_class (
  generic tmpl_var, generic cvr, generic F, generic T, generic body)
{
  generic x  = symbol ("x");
  generic h  = symbol ("helper");

  // the contents of the converter structure
  generic tx = Type (x, F);
  generic f  = Static (Inline (Function (tx, T)));
  generic b  = Define (Type (h, f), body);
  generic val= Struct (b);

  // the converter class with helper routine
  generic var= Specialize (cvr, T, F);
  generic tp = Forall (Type (tmpl_var, MMX_CLASS), MMX_CLASS);
  if (N(tmpl_var) == 2) {
    if ((F == symbol ("F")) && (T == symbol ("T"))) var= cvr;
    tp = Forall (Type (tmpl_var[0], MMX_CLASS),
		 Forall (Type (tmpl_var[1], MMX_CLASS),
			 MMX_CLASS));
  }
  return Define (Type (var, tp), val);
}

void
extender::produce_geninst () {
  generate (make_blackbox_struct (geninst ()));

  generic x  = symbol ("x");
  generic h  = symbol ("helper");
  generic cv = symbol ("convert");
  generic cvr= symbol ("converter");
  generic Cv = symbol ("cast");
  generic Cvr= symbol ("caster");
  generic F  = symbol ("F");
  generic T  = symbol ("T");

  // The converter meta template
  generic b0= Begin (Return (Transtype (x, T)));
  generate (converter_class (list (T, F), cvr, F, T, b0));

  // From abstract to concrete
  generic a1= Transtype (Dot (x, symbol ("rep")), Pointer (T));
  generic b1= Begin (Return (Unpointer (Brackets (a1))));
  generate (converter_class (T, cvr, geninst (), T, b1));

  // From concrete to abstract
  generic b2= Begin (Return (list (geninst (), New (F, x))));
  generate (converter_class (F, cvr, F, geninst (), b2));

  // From abstract aliases to concrete aliases
  generate (converter_class (T, cvr, Alias (geninst ()), Alias (T), b1));

  // From concrete aliases to abstract aliases
  generic a4= New (geninst (), Unalias (x));
  generic b4= Begin (Return (Unpointer (a4)));
  generate (converter_class (F, cvr, Alias (F), Alias (geninst ()), b4));

  // The main conversion routine
  generic cc= Scope (list (cvr, T, F), h);
  generic ff= Inline (Function (Type (x, F), T));
  generic tt= Forall (Type (T, MMX_CLASS), Forall (Type (F, MMX_CLASS), ff));
  generic bb= Return (list (cc, x));
  generate (Define (Type (cv, tt), Begin (bb)));

  // Crude casting meta template
  generic B0= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, F, T, B0));

  // Crude casting for aliases meta template
  generic B1= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, Alias (F), Alias (T), B1));

  // The main conversion routine for crude casting
  generic Cc= Scope (list (Cvr, T, F), h);
  generic Bb= Return (list (Cc, x));
  generate (Define (Type (Cv, tt), Begin (Bb)));
}

list
extender::abstract (int nr, list args, list types) {
  // Convert to abstract type when necessary
  if (nil (args)) return args;
  generic x= car (args);
  generic stype= car (types) [2];
  generic dtype= substitute (stype, geninst (nr), geninst ());
  if (dtype != stype)
    x= list (Specialize (symbol ("cast"), dtype, stype), x);
  return cons (x, abstract (nr, cdr (args), cdr (types)));
}

void
extender::geninst_methods (int nr, generic g) {
  if (!is_func (g, MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1]))
    error (g[1], "Expected a symbol");
  else if (!is_func (g[2], MMX_FUNCTION, 2))
    error (g[2], "Expected a function type");
  else if (!is_func (g[2][1], MMX_CROSS))
    error (g[2][1], "Expected a cross type");
  else {
    g= name_parameters (g);
    generic name = genmethod (g[1]);
    generic fun  = Arrow (curtype (1), name);
    list    args = arg_list (1, N(g[2][1]));
    generic appl = cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    generic dtype= g[2][2];
    generic stype= substitute (dtype, geninst (nr), geninst ());
    if (dtype != stype)
      appl= list (Specialize (symbol ("cast"), dtype, stype), appl);
    if (dtype == MMX_VOID) generate (Define (g, Begin (appl)));
    else generate (Define (g, Begin (Return (appl))));
  }
}

void
extender::geninst_methods (int nr, list l) {
  if (!nil (l)) {
    geninst_methods (nr, car (l));
    geninst_methods (nr, cdr (l));
  }
}

void
extender::produce_geninst (int nr) {
  generate (Type (curtype (nr), Static (gentype)));
  generate (make_blackbox_struct (geninst (nr)));
  list l= specialize (imports, geninst (nr), false);
  l= decl_as_ccpp (l);
  geninst_methods (nr, l);
}

} // namespace mmx
