
/******************************************************************************
* MODULE     : extend.cc
* DESCRIPTION: Automatically generate C++ extension code for generic objects
*              This should later be used in combination with mmx-glue
*              in order to produce glue code for templates
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
#include <quantor.hh>
#include <converter.hh>
#include <error.hh>
namespace mmx {

static converter cvt;

/******************************************************************************
* Error handling and code generation
******************************************************************************/

void
extender::error (generic where, string message) {
  cerr << std_error (message, where);
  nr_errors++;
  /*
  cerr << "Error] ";
  raw_print (cerr, message);
  cerr << lf;
  cerr << "In   ] " << where << lf;
  */
}

void
extender::warning (generic where, string message) {
  cerr << std_warning (message, where);
  /*
  cerr << "Warning] ";
  raw_print (cerr, message);
  cerr << lf;
  cerr << "In     ] " << where << lf;
  */
}

void
extender::generate (generic g) {
  output= cons (g, output);
}

/******************************************************************************
* Useful subroutines
******************************************************************************/

void
nice_print (generic text, list l) {
  cout << text << stroke << lf << indent;
  while (!nil (l)) {
    pretty_print (cout, car (l));
    cout << stroke << lf;
    l= cdr (l);
  }
  cout << unindent << "  " << lf;
}

void
nice_print (symbol text, list l) {
  nice_print ((generic) text, l);
}

generic
make_struct (generic name, list args, list inherit) {
  generic cl= MMX_CLASS;
  if (!nil (inherit)) cl= cons (CCPP_INHERIT, inherit);
  return Define (Type (name, cl), cons (CCPP_STRUCT, args));
}

generic
make_blackbox_struct (generic name) {
  generic rep  = symbol ("rep");
  generic rep2 = symbol ("rep2");
  generic nilv = symbol ("NULL");
  generic bb   = symbol ("bb");
  generic bbr  = Dot (bb, rep);
  generic vp   = Pointer (MMX_VOID);
  generic trep = Type (rep, vp);
  generic trep2= Type (rep2, vp);
  generic tbb  = Type (bb, Const (Alias (name)));
  // FIXME: the constructor without arguments is not nice, but it seems
  // to be incorrectly required by certain template copy operators
  generic r1= Define (Type (name, Inline (Constructor ())),
		      Initialize (generic (list (rep, nilv)), Begin ()));
  generic r2= Define (Type (name, Inline (Constructor (trep2))),
		      Initialize (generic (list (rep, rep2)), Begin ()));
  generic r3= Define (Type (name, Inline (Constructor (tbb))),
		      Initialize (generic (list (rep, bbr)), Begin ()));
  generic r4= Define (Type (name, Inline (Destructor ())),
		      Begin ());
  return make_struct (name, list (trep, r1, r2, r3, r4));
}

/******************************************************************************
* Naming parameters in function types
******************************************************************************/

symbol
parameter (int i) {
  return string ("arg_") * as_string (i);
}

static list
name_parameters (list l, int i) {
  if (nil (l)) return l;
  else return cons (Type (parameter (i), car (l)),
		    name_parameters (cdr (l), i+1));
}

generic
name_parameters (generic g) {
  // name parameters in function type
  if (is_func (g, MMX_TYPE, 2))
    return list (g[0], g[1], name_parameters (g[2]));
  else if (is_func (g, MMX_FUNCTION, 2))
    return list (g[0], name_parameters (g[1]), g[2]);
  else if (is_func (g, MMX_CROSS))
    return cons (g[0], name_parameters (cdr (as_list (g)), 1));
  else return g;
}

/******************************************************************************
* Names
******************************************************************************/

generic
extender::contype (generic name) {
  return symbol (the_prefix * "_") * as_symbol (name);
}

generic
extender::contype_rep (generic name) {
  return symbol (the_prefix * "_") * as_symbol (name) * symbol ("_rep");
}

generic
extender::genmethod (generic name) {
  return symbol ("gen_") * ccpp_protect (name);
}

generic
extender::geninst () {
  return symbol (the_prefix * "_inst");
}

generic
extender::geninst (int nr) {
  return symbol (the_prefix * "_inst_") * symbol (as_string (nr));
}

generic
extender::curtype (int nr) {
  return symbol (the_prefix * "_cur_") * symbol (as_string (nr));
}

/******************************************************************************
* Interface
******************************************************************************/

void
extender::produce () {
  int i;
  produce_geninst ();
  produce_gentype ();
  for (i=1; i<=max_type_args; i++)
    produce_geninst (i);
  produce_contype (types);
}

bool
extension_code (ostream& out, string prefix, generic prog, bool glue_flag) {
  extender ext (prefix);
  ext.parse (prog);
  ext.produce ();
  raw_print (out, ext.preamble);
  out << lf;
  ccpp_print (out, cons (MMX_BEGIN, reverse (ext.output)));
  if (glue_flag) {
    out << "\n";
    out << "#include \"config.hh\"\n";
    out << "#include \"generic.hh\"\n";
    out << "#include \"list.hh\"\n";
    out << "#include \"string.hh\"\n";
    out << "#include \"gentable.hh\"\n";
    out << "#include \"table.hh\"\n";
    out << "#include \"symbol.hh\"\n";
    out << "#include \"mirror.hh\"\n";
    out << "#include \"value.hh\"\n";
    out << "#include \"symbols.hh\"\n";
    out << "#include \"extern.hh\"\n";
    out << "#include \"glued.hh\"\n";
    out << lf;
    ext.output= list ();
    ext.produce_glue ();
    ccpp_print (out, cons (MMX_BEGIN, reverse (ext.output)));
  }
  out << lf;
  raw_print (out, ext.postamble);
  return ext.nr_errors > 0;
}

} // namespace mmx
