%{

/******************************************************************************
* MODULE     : parser.y / parser.cc
* DESCRIPTION: The mathemagix parser
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <symbols.hh>
#include <literal.hh>
#include <lexer.hh>
#include <parser.hh>
#include <error.hh>
namespace mmx {

#define malloc GC_MALLOC
#define free(x)

#define YYSTYPE generic
#define yyparse parse
#define yylex lex
#define yyerror parse_error
#define yylval parse_lval
#define yychar parse_char
#define yydebug parse_debug
#define yynerrs parse_nerrs

static generic parse_result;
static list    parse_warnings;
static list    parse_errors;
static void parse_warning (char* message, generic where);
static void parse_error (char* message);

/******************************************************************************
* Generation of expressions
******************************************************************************/

static generic if_then (generic g_if, generic g_cond, generic g_arg);
static generic define (generic g_assign, generic g_var, generic g_expr);
static generic assign (generic g_assign, generic g_var, generic g_expr);
static generic quantify (generic g_qf, generic g_vars, generic g_body);
static generic sequence (generic g_op, generic g_arg);
static generic htuple (generic g_args);
static generic hvlist (generic g_args);

%}

/******************************************************************************
* One shift/reduce conflict expected (dangling else)
******************************************************************************/

%expect 1 /* dangling else conflict */

/******************************************************************************
* The tokens
******************************************************************************/

%token IDENTIFIER INT FLOAT STRING

%token ADD BREAK CASE CATCH CONTINUE DO ELSE EXPORT EXTEND FOR FROM
%token GENERATE HAS HIDDEN HOLDS IF IMPORT IN INHERIT INTERACTIVE
%token JOIN LOOP OPERATOR POSTFIX RAISE RETURN SPLIT STEP THEN TRY UNTIL
%token VALUE WHILE WITH YIELD MACRO

%token ABSTRACT CONCRETE CONSTANT DIRECT EVOLUTIVE EXTERN INDIRECT
%token INLINE INTERN LOCKED METHOD PRIVATE PROTECTED OUTLINE
%token PACKED PUBLIC UNPACKED VARIABLE

%token DEFINE ASSIGN PLUS_ASSIGN MINUS_ASSIGN TIMES_ASSIGN OVER_ASSIGN
%token TYPE
%token INTO MAPSTO
%token LEFT_FLUX LEFT_FLUX_VAR LEFT_FLUX_STR RIGHT_FLUX
%token FORALL EXISTS
%token WHERE
%token IMPLIES EQUIVALENT
%token SEQOR OR XOR
%token SEQAND AND
%token NOT
%token EQUAL LESS LEQ GREATER GEQ
%token NOT_EQUAL NOT_LESS NOT_LEQ NOT_GREATER NOT_GEQ
%token CAST
%token RANGE
%token PLUS MINUS OPLUS OMINUS
%token TIMES OVER OTIMES OOVER COMPOSE
%token POWER
%token INC DEC SIZE
%token ACCESS
%token BACKPRIME PRIME
%token THIS FOREIGN

%%

/******************************************************************************
* Main production
******************************************************************************/

Program : Cl                     { parse_result= cons (MMX_BEGIN, $1); }
        ;

/******************************************************************************
* Command lists
******************************************************************************/

Sep     : ';'                    { $$ = list (); }
        | ';' Sep                { $$ = list (); }
        ;

Cl      :                        { $$ = list (); }
        | C0 Cl                  { $$ = cons ($1, $2); }
        ;

/******************************************************************************
* Commands
******************************************************************************/

C0      : error Sep              { yyerrok; $$ = MMX_ERROR; }
        | E1 Sep                 { $$ = $1; }
        | '{' Cl '}'             { $$ = sequence (symbol ("begin!", $1), $2);
	                           $$ = extend ($$, $3); }
        | IDENTIFIER '{' Cl '}'  { $$ = sequence (symbol ("begin!", $2), $3);
	                           $$ = list ($1, $$);
	                           $$ = extend ($$, $4); }
        | IF E0 THEN C0          { $$ = if_then ($1, $2, $4); }
        | IF E0 THEN C0 ELSE C0  { $$ = list ($1, $2, $4, $6); }
        | GENERATE E0 DO C0      { $$ = list ($1, $2, $4); }
        | CASE E0
          FROM '{' Casel '}'     { $$ = list ($1, $2, $5); }
        | Conds LOOP C0          { $$ = list ($2, $1, $3); }
        | Conds DO C0            { $$ = list (symbol ("loop!", $2), $1, $3); }
        | SPLIT '{' Splitl '}'   { $$ = cons ($1, $3); }
        | RETURN E1 Sep          { $$ = list ($1, $2); }
        | YIELD E1 Sep           { $$ = list ($1, $2); }
        | RETURN Sep             { $$ = list ($1); }
        | YIELD Sep              { $$ = list ($1); }
        | BREAK Sep              { $$ = list ($1); }
        | CONTINUE Sep           { $$ = list ($1); }
        | DeclMod C0             { $$ = as_list ($1) * list ($2); }
        | Qf '(' Es ')' C0       { $$ = quantify ($1, $3, $5); }
        | F0 DEFINE C0           { $$ = define ($2, $1, $3); }
        | F0 ASSIGN C0           { $$ = assign ($2, $1, $3); }
        | F0 PLUS_ASSIGN C0      { $$ = list ($2, $1, $3); }
        | F0 MINUS_ASSIGN C0     { $$ = list ($2, $1, $3); }
        | F0 TIMES_ASSIGN C0     { $$ = list ($2, $1, $3); }
        | F0 OVER_ASSIGN C0      { $$ = list ($2, $1, $3); }
        ;

Conds   :                        { $$ = list (); }
        | FOR E0 Conds           { $$ = cons (list ($1, $2), $3); }
        | FOR E0 IN E0 Conds     { $$ = cons (list (symbol ("forin!", $1),
						    $2, $4), $5); }
        | WHILE E0 Conds         { $$ = cons (list ($1, $2), $3); }
        | UNTIL E0 Conds         { $$ = cons (list ($1, $2), $3); }
        | STEP E0 Conds          { $$ = cons (list ($1, $2), $3); }
        ;

Casel   :                        { $$ = list (); }
        | ELSE C0                { $$ = list (generic (list ($1, $2))); }
        | Among DO C0 Casel      { $$ = cons (list ($1, $3), $4); }
        ;

Among   : T0                     { $$ = list ($1); }
        | T0 ',' Among           { $$ = cons ($1, $3); }
        ;

Splitl  :                        { $$ = list (); }
        | CASE C0 Splitl         { $$ = cons ($2, $3); }
        ;

/******************************************************************************
* Expressions
******************************************************************************/

E0      : E1                     { $$ = $1; }
        | IF E0 THEN E0 ELSE E0  { $$ = list ($1, $2, $4, $6); }
        | Qf '(' Es ')' E0       { $$ = quantify ($1, $3, $5); }
        | F0 DEFINE E0           { $$ = define ($2, $1, $3); }
        | F0 ASSIGN E0           { $$ = assign ($2, $1, $3); }
        | F0 PLUS_ASSIGN E0      { $$ = list ($2, $1, $3); }
        | F0 MINUS_ASSIGN E0     { $$ = list ($2, $1, $3); }
        | F0 TIMES_ASSIGN E0     { $$ = list ($2, $1, $3); }
        | F0 OVER_ASSIGN E0      { $$ = list ($2, $1, $3); }
        ;

E1      : F0                     { $$ = $1; }
        | Lflux LEFT_FLUX F0     { $$ = list ($2, $1, $3); }
        | Lflux LEFT_FLUX_VAR F0 { $$ = list ($2, $1, $3); }
        | Lflux LEFT_FLUX_STR F0 { $$ = list ($2, $1, $3); }
        | F0 RIGHT_FLUX Rflux    { $$ = list ($2, $1, $3); }
        ;

Lflux   : F0                     { $$ = $1; }
        | Lflux LEFT_FLUX F0     { $$ = list ($2, $1, $3); }
        | Lflux LEFT_FLUX_VAR F0 { $$ = list ($2, $1, $3); }
        | Lflux LEFT_FLUX_STR F0 { $$ = list ($2, $1, $3); }
        ;

Rflux   : F0                     { $$ = $1; }
        | F0 RIGHT_FLUX Rflux    { $$ = list ($2, $1, $3); }
        ;

/******************************************************************************
* Expression lists
******************************************************************************/

Ht      : E0                     { $$= list ($1); }
        | E0 ',' Ht              { $$= cons ($1, $3); }
        ;

Vt      : Ht                     { $$= list (htuple ($1)); }
        | Ht ';' Vt              { $$= cons (htuple ($1), $3); }
        ;

Es      :                        { $$= list (); }
        | Vt                     { $$= hvlist ($1); }
        ;

/******************************************************************************
* Logical formulas
******************************************************************************/

F0      : F1                     { $$ = $1; }
        ;

F1      : F2                     { $$ = $1; }
        | F2 IMPLIES F2          { $$ = list ($2, $1, $3); }
        | F2 EQUIVALENT F2       { $$ = list ($2, $1, $3); }
        ;

F2      : F3                     { $$ = $1; }
        | F3 SEQOR Seqors        { $$ = list ($2, $1, $3); }
        | F3 OR Ors              { $$ = list ($2, $1, $3); }
        | F3 XOR Xors            { $$ = list ($2, $1, $3); }
        ;

Seqors  : F3                     { $$ = $1; }
        | F3 SEQOR Seqors        { $$ = list ($2, $1, $3); }
        ;

Ors     : F3                     { $$ = $1; }
        | F3 OR Ors              { $$ = list ($2, $1, $3); }
        ;

Xors    : F3                     { $$ = $1; }
        | F3 XOR Xors            { $$ = list ($2, $1, $3); }
        ;

F3      : F4                     { $$ = $1; }
        | F4 SEQAND Seqands      { $$ = list ($2, $1, $3); }
        | F4 AND Ands            { $$ = list ($2, $1, $3); }
        ;

Seqands : F4                     { $$ = $1; }
        | F4 SEQAND Seqands      { $$ = list ($2, $1, $3); }
        ;

Ands    : F4                     { $$ = $1; }
        | F4 AND Ands            { $$ = list ($2, $1, $3); }
        ;

F4      : T0                     { $$ = $1; }
        | T0 EQUAL T0            { $$ = list ($2, $1, $3); }
        | T0 LESS T0             { $$ = list ($2, $1, $3); }
        | T0 LEQ T0              { $$ = list ($2, $1, $3); }
        | T0 GREATER T0          { $$ = list ($2, $1, $3); }
        | T0 GEQ T0              { $$ = list ($2, $1, $3); }
        | T0 NOT_EQUAL T0        { $$ = list ($2, $1, $3); }
        | T0 NOT_LESS T0         { $$ = list ($2, $1, $3); }
        | T0 NOT_LEQ T0          { $$ = list ($2, $1, $3); }
        | T0 NOT_GREATER T0      { $$ = list ($2, $1, $3); }
        | T0 NOT_GEQ T0          { $$ = list ($2, $1, $3); }
        | T0 TYPE T0             { $$ = list ($2, $1, $3); }
        | T0 HAS T0              { $$ = list ($2, $1, $3); }
        | T0 HOLDS T0            { $$ = list ($2, $1, $3); }
        ;

/******************************************************************************
* Terms
******************************************************************************/

T0      : T1                     { $$ = $1; }
        | T0 CAST T1             { $$ = list ($2, $1, $3); }
        ;

T1      : T2                     { $$ = $1; }
        | T1 INTO T2             { $$ = list ($2, $1, $3); }
        ;

T2      : T3                     { $$ = $1; }
        | T3 RANGE T3            { $$ = list ($2, $1, $3); }
        ;

T3      : T4                     { $$ = $1; }
        | T3 PLUS T4             { $$ = list ($2, $1, $3); }
        | T3 MINUS T4            { $$ = list ($2, $1, $3); }
        | T3 OPLUS T4            { $$ = list ($2, $1, $3); }
        | T3 OMINUS T4           { $$ = list ($2, $1, $3); }
        ;

T4      : T5                     { $$ = $1; }
        | T4 TIMES T5            { $$ = list ($2, $1, $3); }
        | T4 OVER T5             { $$ = list ($2, $1, $3); }
        | T4 OTIMES T5           { $$ = list ($2, $1, $3); }
        | T4 OOVER T5            { $$ = list ($2, $1, $3); }
        | T4 COMPOSE T5          { $$ = list ($2, $1, $3); }
        ;

T5      : R0                     { $$ = $1; }
        | NOT T5                 { $$ = list ($1, $2); }
        | MINUS T5               { $$ = list ($1, $2); }
        | OMINUS T5              { $$ = list ($1, $2); }
        | INC T5                 { $$ = list ($1, $2); }
        | DEC T5                 { $$ = list ($1, $2); }
        | SIZE T5                { $$ = list ($1, $2); }
        | TypeMod T5             { $$ = list ($1, $2); }
        ;

/******************************************************************************
* Radical terms (twice, because of 'sin x' like function application)
******************************************************************************/

R0      : R1                     { $$ = $1; }
        | R1 S0                  { $$ = list ($1, $2); }
        ;

R1      : R2                     { $$ = $1; }
        | R2 POWER R1            { $$ = list ($2, $1, $3); }
        ;
 
R2      : R3                     { $$ = $1; }
        | R2 INC                 { $$ = list (symbol (".++", $2), $1); }
        | R2 DEC                 { $$ = list (symbol (".--", $2), $1); }
        | R2 NOT                 { $$ = list (symbol (".!", $2), $1); }
        | R2 PRIME               { $$ = list ($2, $1); }
        | R2 BACKPRIME           { $$ = list ($2, $1); }
        | R2 ACCESS R3           { $$ = list ($2, $1, $3); }
        | R2 '(' Es ')'          { $$ = extend (cons ($1, $3), $4); }
        | R2 '[' Es ']'          { $$ = cons (symbol (".[]", $2), $1, $3);
	                           $$ = extend ($$, $4); }
        | SPLIT '(' Es ')'       { $$ = cons ($1, $3); }
        | JOIN '(' E0 ')'        { $$ = list ($1, $3); }
        ;

R3      : S3                     { $$ = $1; }
        | '(' Es ')'             { $$= htuple ($2);
	                           if (N($2)!=1)
				     $$= extend (extend ($$, $1), $3); }
        | '[' Es ']'             { $$ = cons (symbol ("[]", $1), $2);
	                           $$ = extend ($$, $3); }
        ;

S0      : S1                     { $$ = $1; }
        | S1 S0                  { $$ = list ($1, $2); }
        ;

S1      : S2                     { $$ = $1; }
        | S2 POWER S1            { $$ = list ($2, $1, $3); }
        ;

S2      : S3                     { $$ = $1; }
        | S2 INC                 { $$ = list (symbol (".++", $2), $1); }
        | S2 DEC                 { $$ = list (symbol (".--", $2), $1); }
        | S2 NOT                 { $$ = list (symbol (".!", $2), $1); }
        | S2 PRIME               { $$ = list ($2, $1); }
        | S2 BACKPRIME           { $$ = list ($2, $1); }
        | S2 ACCESS S3           { $$ = list ($2, $1, $3); }
        | S2 '(' Es ')'          { $$ = extend (cons ($1, $3), $4); }
        | S2 '[' Es ']'          { $$ = cons (symbol (".[]", $2), $1, $3);
	                           $$ = extend ($$, $4); }
        | SPLIT '(' Es ')'       { $$ = cons ($1, $3); }
        | JOIN '(' E0 ')'        { $$ = list ($1, $3); }
        ;

S3      : Id                     { $$ = $1; }
        | INT                    { $$ = literal (as_symbol ($1),
						 LITERAL_INT_TYPE); }
        | FLOAT                  { $$ = literal (as_symbol ($1),
						 LITERAL_FLOAT_TYPE); }
        | STRING                 { $$ = literal (as_symbol ($1),
						 LITERAL_STRING_TYPE); }
        ;

/******************************************************************************
* Modifiers
******************************************************************************/

DeclMod : EVOLUTIVE              { $$ = list ($1); }
        | EXTERN                 { $$ = list ($1); }
        | INLINE                 { $$ = list ($1); }
        | INTERN                 { $$ = list ($1); }
        | LOCKED                 { $$ = list ($1); }
        | METHOD                 { $$ = list ($1); }
        | PRIVATE                { $$ = list ($1); }
        | PROTECTED              { $$ = list ($1); }
        | OUTLINE                { $$ = list ($1); }
        | PUBLIC                 { $$ = list ($1); }
        | IMPORT                 { $$ = list ($1); }
        | EXPORT                 { $$ = list ($1); }
        | FOREIGN STRING         { $$ = list ($1, $2); }
        ;

TypeMod : ABSTRACT               { $$ = $1; }
        | CONCRETE               { $$ = $1; }
        | CONSTANT               { $$ = $1; }
        | DIRECT                 { $$ = $1; }
        | INDIRECT               { $$ = $1; }
        | PACKED                 { $$ = $1; }
        | UNPACKED               { $$ = $1; }
        | VARIABLE               { $$ = $1; }
        ;

/******************************************************************************
* Identifiers
******************************************************************************/

Id      : IDENTIFIER             { $$ = $1; }
        | OPERATOR Operator      { $$ = $2; }
        | POSTFIX Postfix        { $$ = $2; }
        | THIS                   { $$ = $1; }
        ;

Qf      : FORALL                 { $$ = $1; }
        | EXISTS                 { $$ = $1; }
        ;

Operator: SIZE                   { $$ = $1; }
        | NOT                    { $$ = $1; }
        | INC                    { $$ = $1; }
        | DEC                    { $$ = $1; }
        | BACKPRIME              { $$ = $1; }
        | '[' ']'                { $$ = symbol (".[]", list ($1, $2)); }
        | DEFINE                 { $$ = $1; }
        | ASSIGN                 { $$ = $1; }
        | PLUS_ASSIGN            { $$ = $1; }
        | MINUS_ASSIGN           { $$ = $1; }
        | TIMES_ASSIGN           { $$ = $1; }
        | OVER_ASSIGN            { $$ = $1; }
        | LEFT_FLUX              { $$ = $1; }
        | LEFT_FLUX_VAR          { $$ = $1; }
        | LEFT_FLUX_STR          { $$ = $1; }
        | RIGHT_FLUX             { $$ = $1; }
        | SEQOR                  { $$ = $1; }
        | OR                     { $$ = $1; }
        | XOR                    { $$ = $1; }
        | SEQAND                 { $$ = $1; }
        | AND                    { $$ = $1; }
        | IMPLIES                { $$ = $1; }
        | EQUIVALENT             { $$ = $1; }
        | EQUAL                  { $$ = $1; }
        | LESS                   { $$ = $1; }
        | LEQ                    { $$ = $1; }
        | GREATER                { $$ = $1; }
        | GEQ                    { $$ = $1; }
        | NOT_EQUAL              { $$ = $1; }
        | NOT_LESS               { $$ = $1; }
        | NOT_LEQ                { $$ = $1; }
        | NOT_GREATER            { $$ = $1; }
        | NOT_GEQ                { $$ = $1; }
        | RANGE                  { $$ = $1; }
        | PLUS                   { $$ = $1; }
        | MINUS                  { $$ = $1; }
        | OPLUS                  { $$ = $1; }
        | OMINUS                 { $$ = $1; }
        | TIMES                  { $$ = $1; }
        | OVER                   { $$ = $1; }
        | OTIMES                 { $$ = $1; }
        | OOVER                  { $$ = $1; }
        | COMPOSE                { $$ = $1; }
        | POWER                  { $$ = $1; }
        ;

Postfix : NOT                    { $$ = symbol (".!", $1); }
        | INC                    { $$ = symbol (".++", $1); }
        | DEC                    { $$ = symbol (".--", $1); }
        ;

%%

/******************************************************************************
* Finalize certain constructions
******************************************************************************/

static generic
if_then (generic g_if, generic g_cond, generic g_arg) {
  // Produce dangling else warning if necessary
  generic g= list (g_if, g_cond, g_arg);
  if ((N (g_arg) == 4) && (car (g_arg) == MMX_IF)) {
    generic g_else = g_arg[3];
    parse_warning ("dangling else", g_else);
  }
  return g;
}

static generic
define (generic g_assign, generic g_var, generic g_expr) {
  // Transforms a declaration like
  //   f(...)...(...):R == ... or
  //   f(...)...(...):R := ...
  // into a declaration var:funtype ==/:= lambdaexpr

  if (!is_func (g_var, MMX_TYPE))
    return list (g_assign, g_var, g_expr);

  generic val = g_expr;
  generic var = g_var[1];
  generic type= g_var[2];
  while (is_list (var)) {
    generic dom= cons (MMX_CROSS, flatten_types (cdr (as_list (var))));
    val = list (MMX_LAMBDA, cdr (as_list (var)), type, val);
    type= list (MMX_FUNCTION, dom, type);
    var = var[0];
  }
  return list (g_assign, list (MMX_TYPE, var, type), val);
}

static generic
assign (generic g_assign, generic g_var, generic g_expr) {
  // Use :== symbol for declarations of variables so as
  // to distinguish between declarations and assignments
  if (!is_func (g_var, MMX_TYPE))
    return list (g_assign, g_var, g_expr);
  return define (symbol (":==", g_assign), g_var, g_expr);
}

static bool
is_declaration (generic g) {
  if ((N(g) == 3) && ((g[0] == MMX_DEFINE) || (g[0] == MMX_DEFINEALIAS)))
    if ((N(g[1]) == 3) && (g[1][0] == MMX_TYPE)) return true;
  return false;
}

static generic
quantify (generic g_qf, generic g_vars, generic g_body) {
  // Replace quantifiers over multiple variables by
  // several quantifiers over one variable and
  // rewrite declaration bodies
  list l= as_list (g_vars);
  if (nil (l)) return g_body;
  else {
    generic body= quantify (g_qf, cdr (l), g_body);
    if (is_declaration (body)) {
      generic var = body[1][1];
      generic type= list (g_qf, car (l), body[1][2]);
      generic val = list (g_qf == MMX_FORALL? MMX_FORALLVAL: MMX_EXISTSVAL,
			  car (l), type, body[2]);
      return list (body[0], list (body[1][0], var, type), val);
    }
    else return list (g_qf, car (l), body);
  }
}

static generic
sequence (generic g_op, generic g_arg) {
  // Isolate declarations in command blocks
  if (nil (g_arg)) return list (g_op);
  if (car (g_arg) == MMX_ERROR)
    return sequence (g_op, cdr (g_arg));

  generic item= car (g_arg);
  generic next= sequence (g_op, cdr (g_arg));
  generic head= car (next);
  generic tail= cdr (next);

  if (is_declaration (item)) {
    if (head == MMX_LOCAL)
      return cons (symbol ("local!", g_op), item, tail);
    return list (symbol ("local!", g_op), item, next);
  }
  else {
    if (head == MMX_BEGIN)
      return cons (symbol ("begin!", g_op), item, tail);
    return list (symbol ("begin!", g_op), item, next);
  }

  fatal_error ("unexpected situation", "sequence", "parser.y");
  exit (1);
}

static generic
htuple (generic g_args) {
  // produce ,-separated tuple
  if (N(g_args) == 1) return car (g_args);
  else return cons (symbol ("tuple!", g_args), as_list (g_args));
}

static generic
hvlist (generic g_args) {
  // produce ;-separated tuple of ,-separated tuples and
  // return as argument list
  if (N(g_args) == 1) {
    generic ht= car (g_args);
    if (is_func (ht, "tuple!")) return cdr (ht);
    else return list (ht);
  }
  else {
    generic vt= cons (symbol ("vtuple!", g_args), as_list (g_args));
    return list (vt);
  }
}

/******************************************************************************
* Exported parsing functions
******************************************************************************/

void
parse_warning (char* s, generic where) {
  generic war= std_warning (string (s), where);
  parse_errors= cons (war, parse_errors);
}

void
parse_error (char* s) {
  int pos1= lex_origin + lex_prev;
  int pos2= lex_origin + lex_pos;
  generic err= std_error (string (s), symbol ("", pos1, pos2));
  parse_errors= cons (err, parse_errors);
}

generic
parse (string s, list& errors) {
  start_lexer (s);
  parse_result= generic ();
  parse_errors= list ();
  parse ();
  errors= parse_errors;
  return parse_result;
}

} // namespace mmx
